
(* 
  *$Id: action.ml,v 1.15 2002/09/02 18:54:45 emmanuel Exp $ 
*)

(* 

  1/ a partir des objectifs et des actions locales precedents on : 

    calcul d'amelioration d'objectifs
 
    teste si on est seul 
    teste le niveau d'alerte 

    si alone ou tranquille : faire action locale s'il en reste sinon action globale 
    sinon 
      si dense on essaie de sortir de la zone 
      sinon  combat (minimax) 


*)


let print_action s = if !Debug.action then (Format.eprintf "Action:  %s@." s);;
let print_action_c p =  if !Debug.action then (print_action ("("^(string_of_int (fst p))^","^(string_of_int(snd p))^")"));;

type pos = int * int;;
type objectifs =   (int array * (pos list )) ref  ;; 

type action = 
  North  
| South 
| West 
| East 
| Pick of string list * objectifs 
| Drop of string list * objectifs
| Wait
| Next;;
 

let char_to_action c = match c with 
  'n' -> North 
  | 's' -> South
  | 'e' -> East
  | 'w' -> West
  | 'd' -> Drop ([],(ref([||],[])))
  | _ -> Drop ([],(ref([||],[])))
;;



(* true : pas de boucle detectee *)
let situation hist n obj =
  let d,o = obj in 
    if o = [] then true 
    else 
      if List.length hist < n then true 
      else 
        let o1::ol = o in 
        let (x,y) = State.my_location () in 
        let my_d = Distance.get_distance d x y in 
          my_d < List.nth hist n;;
 

(** from GOTO *)
let n_of (x, y) = x, (succ y);;
let e_of (x, y) = (succ x), y ;;
let s_of (x, y) = x, (pred y) ;;
let w_of (x, y) = (pred x), y ;;
		
let infinite = max_int ;;

let dir_chars = [| North; East; South; West |] ;;

let get_dist dist (x,y) = 
  Distance.get_distance dist x y


let next dist xy =
  let adj = [|n_of xy; e_of xy; s_of xy; w_of xy |] in
  let next_xy = ref (n_of xy)  in
  let dir = ref Wait in
    for i = 0 to 3 do
      if get_dist dist adj.(i) <= get_dist dist !next_xy then (
	next_xy := adj.(i);
        dir := dir_chars.(i))
    done;
    (*!next_xy,*) !dir
;;
(** fin de from GOTO *)



let all_bases () = [||],[];;

let rec new_base ()   = 
  let l = !State.homes in 
    match l with 
      | [] -> begin  Hashtbl.iter (fun a (_,x,y) -> State.homes := (x,y) :: !State.homes) State.package_locations; 
	  if !State.homes  <> [] then new_base() else all_bases () end
      | _ -> 
     let (x,y) as pos = State.my_location() in  
     let d = Distance.distance l in
     let p = Distance.plus_proche x y d in 
       begin
	 try
	   let pkgl =  !(Hashtbl.find State.packages_at_location p) 
	   in 
	   if pkgl = [] then begin
	     State.homes := List.filter (fun x -> x <> p) !State.homes;
	     new_base()
	   end
       else begin
	 print_action_c p;
	 (Distance.distance [p]),[p]
       end
	   with _-> d,[p]
       end;;
 
let prochain_objectif obj = 
  let d,lpos = !obj in 
    match lpos with 
      | [] ->  new_base() 
      | _ -> !obj;;


    

let send a b = match a with 
    West -> Communication.send b 'w' []
  | East -> Communication.send b 'e' []
  | North -> Communication.send b 'n' []
  | South -> Communication.send b  's' []
  | Drop (lp,o) -> Communication.send b  'd' lp
  | Pick (lp,o) -> Communication.send b 'p' lp
  | Wait -> Communication.send b 'p' [] (* failwith "action:send:wait" (* Communication.send 1 'd' []*) *)
  | Next -> failwith "action:send:next"



   
let make_action al obj  = match !al with 
  [] -> let d,lpos = !obj in let my_x,my_y = State.my_location () in let a = next d (my_x,my_y) in send a 1; a
  | (b,a)::q -> send a b; al := q;a
;;



let pkg_ids_of_descrs pkg_descrs =
  List.map (fun pkg_descr -> pkg_descr.Select3.id) pkg_descrs

let rec add_not_mem x xs =
  match xs with
      [] -> [x]
    | x'::xs' -> if x=x' then xs else x'::(add_not_mem x xs')

let pkg_dests_of_descrs pkg_descrs =
  let rec loop pds dls =
    match pds with
	[] -> dls
      | pd::pds
	  -> loop pds (add_not_mem (let {State.x=x; State.y=y}=pd.Select3.dest in (x,y)) dls)
  in
    loop pkg_descrs []  

(* boucle principale *)

let select = Select3.main_select;;

let my_packages () = Hashtbl.find State.robot_packages (!State.my_id);;
let package_dests l = List.map (fun x -> Hashtbl.find State.package_dests x) l;;
let package_poss l = List.map (fun {State.x=x; State.y=y; State.weight = w} -> (x,y)) l;;

let rec reduce l = match l with 
  [] -> [] 
  | h::t -> h::(reduce(List.filter (fun x -> x <> h) t));;
  
let update  rob_pkgl score action = 
begin
  match action with 
    | Drop ([],_) -> () (*failwith "action:update:drop:empty"*)
    | Drop (l,obj) -> 
      if score < !State.score then begin
        let d,pl = !obj in
          match pl with 
            | [] -> failwith "action:update:drop"
            | [o1] -> obj:=d,[]
            | o1::os -> let nd = Distance.distance [List.hd os] in obj:= nd, os (* (List.filter (fun x -> x <> o1) os)  *)
      end
    | Pick ([],obj) -> (let d,pl = !obj in match pl with [] -> () 
			  | [o1] -> obj := d,[] 
			  | o1::os  -> let nd = Distance.distance [List.hd os] in obj:= nd, os)
		       (* failwith "action:update:pick:empty" *)
    | Pick (id_l,obj) -> 
        if (List.length (!(my_packages())) > (List.length rob_pkgl)) then  begin
	  let d,pl = !obj in 
          let pos_l = reduce (package_poss ((package_dests id_l))) in 
	    match pl  with 
	      | [] ->   let nd = Distance.distance [(List.hd pos_l)] in  obj:=nd,pos_l
	      | [o1] ->  let nd = Distance.distance [(List.hd pos_l)] in  obj:=nd,pos_l
	      | o1::os -> let nd = Distance.distance [List.hd os] in obj := nd, os@pos_l (* (List.filter (fun x -> x <> o1) os)@pos_l  *)
	    end
    | _ -> ()
end
;;

let c = ref 0;;


let rec loop (obj:objectifs)   al hist  pkgl score action   = 

  if !Debug.action then (incr c;  print_action (string_of_int !c));
  let my_x,my_y = State.my_location() in 

  let all_pack_id = State.get_robot_packages !State.my_id in 
  let all_pack_dest = List.map (fun x-> x,Hashtbl.find State.package_dests x) !all_pack_id in 

  let sit = situation hist 5 !obj in 
  let ev = Alerte.evaluationRisques() in 
  begin 
    update pkgl score action;
    if !Debug.action then begin 
      let d,pos_l = !obj in 
	print_action "POSITIONS";
        List.iter (print_action_c) pos_l;
	print_action "FIN POSITIONS"
    end;
    obj:= prochain_objectif obj;
    let d, ol =  !obj in 
    let objectif_atteint = begin
      match ol with 
	| [] -> false
	| (x1,y1)::_ -> (x1 == my_x) && (y1 == my_y) 
    end
    in 
    match ev,objectif_atteint with 
	(Alerte.Alone,_) 
      | (Alerte.Tranquille,_)
      | (_,true)  ->
	  let d,ol = !obj in 
	    begin
	      print_action (" --> "^(string_of_int (List.length !al))^"   ..  "^(string_of_int (List.length ol)));
	      print_action ("D(10,12)"^(string_of_int (Distance.get_distance d 10 12)));
	      (List.iter print_action_c ol);

	      match (!al),ol with 
		  (* au debut ou quand on ne sait plus quoi faire *)
		 (*  [],[] ->  (let d,p = new_base () in print_action "new_base"; obj := d,p; al := [1,next d (my_x,my_y)]; print_action "ee")*)
		      (* pas d'action locale mais au moins 1 objectif *)
		| [],((x1,y1) as o1)::os -> 
		   
		    (* objectif atteint *)
		    if objectif_atteint then 
		      begin
			let all_pack_here = List.filter ( fun (id,{State.x=x;State.y=y; State.weight=w})  -> (x == my_x) && (y == my_y)) all_pack_dest in 

			  (*  y a-t-il qqchose a' poser *)
			  if all_pack_here <> [] then (	  print_action "qqchose a' poser";
			    let remove_pkg = (List.map (fun (id,_) -> id) all_pack_here) in  
			      al := (1,(Drop ( remove_pkg ,obj))) :: !al ) 
			      

			    (* il n'y a rien a' poser *)
			  else ( 
			    print_action "qq chose a' prendre";
			    let pick_l,drop_l = select [] in 
			    let lid = List.map  (fun (x,y) -> x) pick_l in
			    let flat_lid = List.fold_left (@) [] lid in 
			    let lpos = List.map (fun (x,y) -> y) pick_l in

			      print_action (string_of_int (List.length pick_l));
			      print_action (string_of_int (List.length flat_lid));
			      print_action (string_of_int (List.length lpos));

			    al := [1,Pick (flat_lid,obj)])
		      end
		    else (  print_action "suivre le mouvement"; ()) 


		(* ne rien calculer on fera l'action courante *)      
		| _,_ -> ()
	    end
      | (Alerte.ZoneDense,false) ->  
	  print_action "ZoneDense";
	  let d,pos_l = !obj in begin 
            match pos_l with 
		[] -> let d,p = new_base () in print_action "new_base"; obj := d,p; 
	      | _ -> ()
	    end	  ;
	    let d,pos_l = !obj in 
	    let (nx1,ny1) = List.hd pos_l in 
	  let _,coup = Distance.distance_guepier 3 my_x my_y  nx1 ny1  d  2 2 2 in  al := [1,(char_to_action coup)]
      | (Alerte.CombatLocal (scope, robots)),_ ->  
	  print_action "ZoneCombatLocal";
	  let d,pos_l = !obj in let coup,bid,_ = Alphabeta.find_move  d scope robots in al := [bid,(char_to_action coup)]
end;
  let api = !all_pack_id 
  and score = !State.score in 
  let a = make_action al obj in 
  loop obj al hist api score a 
    
;;




let _ =
  Communication.start ();
  let b = new_base() in 
  loop (ref (b)) (ref []) [] [] 0 Next;;





(* 
			      (match !obj with 
				   d,[] -> obj := d,[]
				 | d,[o1] -> obj := d,[]
				 | d,((o1::ol) as l)  -> let (x,y) = o1 in let d = Distance.distance [o1] in obj:=(d,ol)))

			      ( match lpos with 
				    [] -> obj := d,os
				  | _ -> let d = Distance.distance [List.hd lpos] in   obj := d,  lpos@ os))

*)



