
open State

(****)

let input = ref None
let output = ref None
let get_input () = match !input with Some inp -> inp | None -> assert false

let write_board ch =
  Printf.fprintf ch "%d %d\n" !w !h;
  for i = 1 to !h - 2 do
    Printf.fprintf ch "%s\n" (String.sub !board (!w * i + 1) (!w - 2))
  done

let read_board ch =
  Scanf.fscanf ch "%d %d" (fun w' h' -> w := w' + 2; h := h' + 2);
  board := String.make (!w * !h) '#';
  for i = 1 to !h - 2 do
    Scanf.fscanf ch "%s"
      (fun s -> String.blit s 0 !board (!w * i + 1) (!w - 2))
  done;
  homes := [];
  for x = 1 to !w - 2 do
    for y = 1 to !h - 2 do
      if get_contents x y = '@' then begin
        homes := (x, y) :: !homes;
        set_contents x y '.'
      end
    done
  done;
  if !Debug.comm then write_board stderr;
  if !Debug.comm then begin
    Format.eprintf "Homes: ";
    List.iter (fun (x, y) -> Format.eprintf "%d,%d " x y) !homes;
    Format.eprintf "@."
  end

let read_configuration ch =
  Scanf.fscanf ch "%s %d %d"
    (fun i max mon -> my_id := i; max_weight := max; money := mon);
  if !Debug.comm then Format.eprintf "Robot id: %s@." !my_id;
  if !Debug.comm then Format.eprintf "Max weight: %d@." !max_weight;
  if !Debug.comm then Format.eprintf "Money: %d@." !money

let read_updates () =
  let ch = get_input () in
  if !Debug.comm then Format.eprintf "Server response: ";
  let s = input_line ch in
  let len = String.length s in
  let pos = ref 0 in
  let rec skip_spaces () =
    while
      !pos < len && match s.[!pos] with ' ' | '\t' -> true | _ -> false
    do incr pos done
  in
  let read_int () =
    skip_spaces ();
    let start = !pos in
    while
      !pos < len && match s.[!pos] with '0'..'9' -> true | _ -> false
    do incr pos done;
    if !pos = start then begin
      Format.eprintf "@.Missing number at position %d with score %d in \"%s\".@." !pos !State.score s;
      exit 1
    end;
    String.sub s start (!pos - start)
  in
  let robots = ref [] in
  let current_robot = ref "" in
  let rec decode_action () =
    skip_spaces ();
    if !pos < len then
      let c = s.[!pos] in
      incr pos;
      match c with
        '#' ->
          current_robot := read_int ();
          robots := !current_robot :: !robots;
          if !Debug.comm then Format.eprintf "#%s " !current_robot
      | 'N' ->
          move_robot !current_robot 0 1;
          if !Debug.comm then Format.eprintf "%c " c
      | 'S' ->
          move_robot !current_robot 0 (-1);
          if !Debug.comm then Format.eprintf "%c " c
      | 'E' ->
          move_robot !current_robot 1 0;
          if !Debug.comm then Format.eprintf "%c " c
      | 'W' ->
          move_robot !current_robot (-1) 0;
          if !Debug.comm then Format.eprintf "%c " c
      | 'P' ->
          let id = read_int () in
          robot_pick_package !current_robot id;
          if !Debug.comm then Format.eprintf "P %s " id
      | 'D' ->
          let id = read_int () in
          robot_drop_package !current_robot id;
          if !Debug.comm then Format.eprintf "D %s " id
      | 'X' ->
          let x = int_of_string (read_int ()) in
          if !Debug.comm then Format.eprintf "X %d " x;
          skip_spaces ();
          assert (s.[!pos] = 'Y');
          incr pos;
          skip_spaces ();
          let y = int_of_string (read_int ()) in
          if !Debug.comm then Format.eprintf "Y %d " y;
          set_robot !current_robot x y
      | _ ->
          Format.eprintf "Response not understood at pos %d: %s@."
            (!pos - 1) s;
          exit 1
  in
  while !pos < len do
    decode_action ()
  done;
  if !Debug.comm then Format.eprintf "@.";
  let dead_robots =
    Hashtbl.fold
      (fun id _ dead -> if List.mem id !robots then dead else id :: dead)
      robot_locations []
  in
  List.iter kill_robot dead_robots

let read_packages () =
  let ch = get_input () in
  if !Debug.comm then Format.eprintf "Packages: ";
  let s = input_line ch in
  let len = String.length s in
  let pos = ref 0 in
  let rec skip_spaces () =
    while
      !pos < len && match s.[!pos] with ' ' | '\t' -> true | _ -> false
    do incr pos done
  in
  let read_int () =
    skip_spaces ();
    let start = !pos in
    while
      !pos < len && match s.[!pos] with '0'..'9' -> true | _ -> false
    do incr pos done;
    if !pos = start then begin
      Format.eprintf "@.Missing number at position %d with score %d in \"%s\".@." !pos !State.score s;
      exit 1
    end;
    String.sub s start (!pos - start)
  in
  let current_robot = ref "" in
  let (x', y') = my_location () in
  let rec decode_package () =
    skip_spaces ();
    if !pos < len then
      let id = read_int () in
      let x = int_of_string (read_int ()) in
      let y = int_of_string (read_int ()) in
      let w = int_of_string (read_int ()) in
      set_package id x y x' y' w;
      if !Debug.comm then Format.eprintf "%s %d %d %d " id x y w
  in
  while !pos < len do
    decode_package ()
  done;
  kill_unknown_packages x' y';
  if !Debug.comm then Format.eprintf "@."

let send bid cmd args =
  if !money < bid then begin
    Format.eprintf "No money left!@.";
    exit 1
  end;
  money := !money - abs bid;
  let cmd =
    match cmd with
      'n' -> "Move N"
    | 's' -> "Move S"
    | 'e' -> "Move E"
    | 'w' -> "Move W"
    | 'p' -> "Pick"
    | 'd' -> "Drop"
    | _   -> Format.eprintf "Unknown command %c@." cmd; exit 1
  in
  let args = String.concat " " ("" :: args) in
  let s = Format.sprintf "%d %s%s" bid cmd args in
  if !Debug.comm then Format.eprintf "Sending: \"%s\"@." s;
  let ch = match !output with Some ch -> ch | None -> assert false in
  Printf.fprintf ch "%s\n" s;
  flush ch

let print_map = ref false

let connect_and_read_board () =
  let host_port = ref [] in
  Arg.parse
    [("-debug", Arg.String Debug.set,
      "<module>  debug the given module");
     ("-map", Arg.Unit (fun () -> print_map := true),
      "  print the map at each iteration")]
    (fun s -> host_port := s :: !host_port)
    "Options:";
  let (inp, outp) =
    match !host_port with
      [port; host] ->
        let host = Unix.gethostbyname host in
        let port = int_of_string port in
        Unix.open_connection (Unix.ADDR_INET (host.Unix.h_addr_list.(0), port))
    | [] ->
        (stdin, stdout)
    | _ ->
        Format.eprintf "Wrong number of options.@.";
        exit 1
  in
  input := Some inp;
  output := Some outp;
  output_string outp "Player\n"; flush outp;
  read_board inp

let start () =
  connect_and_read_board ();
  let inp = get_input () in
  read_configuration inp;
  read_updates ();
  if !print_map then
    Print.configuration stdout;
  read_packages ();
  if !print_map then begin
    Print.map stdout;
    Print.me stdout
  end

let send bid cmd args =
  send bid cmd args;
  read_updates ();
  incr moves;
  read_packages ();
  if !print_map then begin
    Print.map stdout;
    Print.me stdout
  end
