open State

let n_of (x, y) = x, (succ y);;
let e_of (x, y) = (succ x), y ;;
let s_of (x, y) = x, (pred y) ;;
let w_of (x, y) = (pred x), y ;;
		
let infinite = max_int ;;

let dir_chars = [| 'n'; 'e'; 's'; 'w' |] ;;

let get_distance dist (x,y) = 
  Distance.get_distance dist x y


let next dist xy =
  let adj = [|n_of xy; e_of xy; s_of xy; w_of xy |] in
  let next_xy = ref xy  in
  let dir = ref ' ' in
    for i = 0 to 3 do
      let new_d = get_distance dist adj.(i) 
      and old_d = get_distance dist !next_xy in
      if (new_d < old_d) || ((new_d = old_d) && ((Random.int 2) = 0)) then (
	next_xy := adj.(i);
        dir := dir_chars.(i))
    done;
    !next_xy, !dir

let list_rem x xs = List.filter (fun x' -> x<>x') xs ;; 

let target_stack = ref ([] : (int * int) list)

type targets_type =
    { 
      mutable deliveries : (int*int) list;
      mutable bases : (int*int) list;
      mutable route : int array
    }

let targets =
  { 
    deliveries = [];
    bases = [];
    route = [||]
  }

let print_targets () =
  Printf.eprintf "Targets: \n"; 
  Printf.eprintf"\n deliveries: ";
  List.iter (fun (x,y) -> Printf.eprintf" (%d,%d);" x y) targets.deliveries; 
  Printf.eprintf"\n base: ";
  List.iter (fun (x,y) -> Printf.eprintf" (%d,%d);" x y) targets.bases;  
  Printf.eprintf"\n";
  flush stdout

let next_dir xy = snd (next targets.route xy)

let next_target () = 
  match targets.deliveries with
      [] ->
	begin
	  match targets.bases with
	      [] -> 
		begin
		  Printf.eprintf"Don't know what to do !\n";
		  print_targets(); flush stdout;
		  failwith"Don't know what to do !" 
		end
	    | _ -> targets.route <- Distance.distance targets.bases
	end
    | _ -> targets.route <- Distance.distance targets.deliveries

let do_rem_bases xy =
  if List.mem xy targets.bases then (
    targets.bases <- list_rem xy targets.bases;
    true )
  else
    false
	  
let packages_to_drop_at xy =
  let my_pkg = !(State.get_robot_packages !State.my_id) in
   List.filter 
     (fun id -> let {x = x; y =y}=Hashtbl.find State.package_dests id 
                in xy = (x,y)) 
     my_pkg

let rec add_not_mem x xs =
  match xs with
      [] -> [x]
    | x'::xs' -> if x=x' then xs else x'::(add_not_mem x xs')

let packages_to_pick_at (x,y) =
  match List.map snd !(State.get_packages_at_location x y) with
      [] -> []
    | pkg_ids ->  Select.select 
		   (State.my_location())
		   pkg_ids
		   !(State.get_robot_packages !State.my_id)
		   !State.max_weight

let pkg_ids_of_descrs pkg_descrs =
  List.map (fun pkg_descr -> pkg_descr.Select.id) pkg_descrs

let rec add_not_mem x xs =
  match xs with
      [] -> [x]
    | x'::xs' -> if x=x' then xs else x'::(add_not_mem x xs')

let pkg_dests_of_descrs pkg_descrs =
  let rec loop pds dls =
    match pds with
	[] -> dls
      | pd::pds
	  -> loop pds (add_not_mem (let {x=x; y=y}=pd.Select.dest in (x,y)) dls)
  in
    loop pkg_descrs []  

let bid() = 1

let send_move dir =
  Communication.send (bid()) dir []

let send_pick pkg_ids =
  Communication.send (bid()) 'p' pkg_ids

let send_drop pkg_ids =
  Communication.send (bid()) 'd' pkg_ids

let move = ref false 

let goto () =
  let xy = State.my_location() in
    if !move then
      begin
	send_move (next_dir xy);
	move := false
      end
    else
      begin
	match packages_to_drop_at xy with
	    [] -> 
	      begin
		match packages_to_pick_at xy with
		    [] -> 
		      begin
			Printf.eprintf"Est-ce ici ?\n"; flush stdout;
			(if do_rem_bases xy then next_target());
			send_move (next_dir xy)
		      end
		  | pkg_descrs -> 
		      begin
			send_pick (pkg_ids_of_descrs pkg_descrs);
			targets.deliveries <- 
		        (pkg_dests_of_descrs pkg_descrs)@targets.deliveries;
			next_target();
			move := true
		      end
	      end
	  | pkg_ids -> 
	      begin
		send_drop pkg_ids;
		targets.deliveries <- list_rem xy targets.deliveries;
		next_target()
	      end
      end
      
		
      
