
let last_line = ref ""

let split_cmd s =
  let pos = ref 0 in
  while s.[!pos] <> ' ' do
    incr pos
  done;
  (int_of_string (String.sub s 0 !pos),
   String.sub s (!pos + 1) (String.length s - !pos - 1))

exception Failed

let decode_command s =
  let pos = ref 0 in
  let len = String.length s in
  let rec skip_spaces () =
    while
      !pos < len && match s.[!pos] with ' ' | '\t' -> true | _ -> false
    do incr pos done
  in
  let on_digit () =
    !pos < len && match s.[!pos] with '-' | '+' | '0'..'9' -> true | _ -> false
  in
  let read_unsigned () =
    let start = !pos in
    while
      !pos < len && match s.[!pos] with '0'..'9' -> true | _ -> false
    do incr pos done;
    if !pos = start then raise Failed;
    String.sub s start (!pos - start)
  in
  let read_int () =
    match s.[!pos] with
      '-' -> incr pos; - int_of_string (read_unsigned ())
    | '+' -> incr pos; int_of_string (read_unsigned ())
    | _   -> int_of_string (read_unsigned ())
  in
  skip_spaces ();
  let bid = if on_digit () then read_int () else 1 in
  skip_spaces ();
  if !pos = len then raise Failed;
  let cmd = s.[!pos] in
  incr pos;
  let args = ref [] in
  while skip_spaces (); !pos < len do
    args := read_unsigned () :: !args
  done;
  (bid, cmd, List.rev !args)

let print_help () =
  print_string "\
Commands are of the form \"10 p 3 4 5\".\n\
The first number is the bid.  If it is ommitted, a value of 1 is used.\n\
The possible commands are:\n\
- n,s,e,w: move in the corresponding direction;
- p: peek some packages (followed by a list of packages);
- d: peek some packages (followed by a list of packages);
- i: print robot and package information;
- m: redisplay the map;
- q: exit;
- h/?: print this help message.
";
  flush stdout

let rec process_commands () =
  let l = read_line () in
  let l = if l = "" then !last_line else l in
  last_line := l;
  begin try
    let (bid, cmd, args) = decode_command l in
    begin match cmd with
      'n' | 's' | 'e' | 'w' ->
        if args <> [] then raise Failed;
        Communication.send bid cmd args
    | 'p' | 'd' ->
        Communication.send bid cmd args
    | 'i' ->
        Print.robots stdout;
        Print.packages stdout
    | 'm' ->
        Print.map stdout;
        Print.me stdout
    | 'h' | '?' ->
        print_help ()
    | 'q' ->
        exit 0
    | _ ->
        raise Failed
    end;
  with Failed ->
    Format.printf "Syntax error@.@.";
    print_help ()
  end;
  process_commands ()

let _ =
  Communication.print_map := true;
  Communication.start ();
  process_commands ()
