
open State

let map ch =
  let map = String.copy !board in
  let get_contents x y = map.[x + y * !w] in
  let set_contents x y c = map.[x + y * !w] <- c in
  Hashtbl.iter
    (fun _ {x = x; y = y} ->
       if get_contents x y = '.' then set_contents x y '!')
    package_dests;
  Hashtbl.iter
    (fun _ (st, x, y) ->
       if get_contents x y = '.' then
         set_contents x y (if st = `Yes then 'p' else '?'))
    package_locations;
  List.iter (fun (x, y) -> set_contents x y '@') !homes;
  Hashtbl.iter
    (fun id (x, y) ->
       let me = id = !my_id in
       begin match get_contents x y with
         '.' | '!' | 'p' | '?' -> set_contents x y (if me then 'M' else 'R')
       | '@'                   -> set_contents x y (if me then 'm' else 'r')
       | 'm' | 'M' | 'r' | 'R' | '*' ->
                                  set_contents x y '*'
       | c                     -> Format.eprintf "???%c???@." c; assert false
       end)
    robot_locations;
  for i = !h - 2 downto 1 do
    Printf.fprintf ch "%s\n" (String.sub map (!w * i + 1) (!w - 2))
  done;
  flush stdout

let me ch =
  let ch = Format.formatter_of_out_channel ch in
  let (x, y) = my_location () in
  Format.fprintf ch
    "Pos: (%d, %d)    Money: %d    Score: %d    Max weight: %d@."
    x y !money !score !max_weight;
  Format.fprintf ch
    "Moves: %d    Time credit: %gs@."
    !moves (Time.credit ());
  let l = !(get_robot_packages !my_id) in
  if l <> [] then begin
    Format.fprintf ch "@[<2>Carried packages:@ ";
    let first = ref true in
    List.iter
      (fun id ->
         if not !first then Format.fprintf ch ";@ "; first := false;
         Format.fprintf ch "%s:" id;
         begin try
           let {x = x; y = y; weight = w} = Hashtbl.find package_dests id in
           Format.fprintf ch " (%d, %d) %dkg" x y w
         with Not_found -> () end)
      l;
    Format.fprintf ch "@]@."
  end;
  let l = !(get_packages_at_location x y) in
  if l <> [] then begin
    Format.fprintf ch "@[<2>Available packages:@ ";
    let first = ref true in
    List.iter
      (fun (_, id) ->
         if not !first then Format.fprintf ch ";@ "; first := false;
         Format.fprintf ch "%s:" id;
         begin try
           let {x = x; y = y; weight = w} = Hashtbl.find package_dests id in
           Format.fprintf ch " -> (%d, %d) %dkg" x y w
         with Not_found -> () end)
      l;
    Format.fprintf ch "@]@."
  end

let configuration ch =
  let ch = Format.formatter_of_out_channel ch in
  Format.fprintf ch "Robot id: %s@." !my_id;
  Format.fprintf ch "Max weight: %d@." !max_weight;
  Format.fprintf ch "Money: %d@." !money

let robots ch =
  let ch = Format.formatter_of_out_channel ch in
  Format.fprintf ch "@[<2>Robots: ";
  let first = ref true in
  Hashtbl.iter
    (fun id (x, y) ->
       if not !first then Format.fprintf ch ";@ "; first := false;
       if id = !my_id then
         Format.fprintf ch "%s(me): " id
       else
         Format.fprintf ch "%s: " id;
      Format.fprintf ch "(%d, %d)" x y)
    robot_locations;
  Format.fprintf ch "@]@."

let packages ch =
  let ch = Format.formatter_of_out_channel ch in
  if PackageSet.is_empty !package_list then
    Format.fprintf ch "No known package.@."
  else begin
    Format.fprintf ch "@[<2>Packages:@ ";
    let first = ref true in
    PackageSet.iter
      (fun id ->
         if not !first then Format.fprintf ch ";@ "; first := false;
         Format.fprintf ch "%s: " id;
         begin try
           let (st, x, y) = Hashtbl.find package_locations id in
           if st = `Yes then
             Format.fprintf ch "(%d, %d)" x y
           else
             Format.fprintf ch "(%d, %d)?" x y
         with Not_found -> try
           let id' = Hashtbl.find package_carried id in
           Format.fprintf ch "(rob %s)" id'
         with Not_found ->
           Format.eprintf "???%s@." id
         end;
         begin try
           let {x = x; y = y; weight = w} = Hashtbl.find package_dests id in
           Format.fprintf ch " -> (%d, %d) %dkg" x y w
         with Not_found -> () end)
      !package_list;
    Format.fprintf ch "@]@."
  end
