
(**
 $Id: select.ml,v 1.5 2002/09/01 10:59:01 emmanuel Exp $
**)


(* choix d'un ensemble de package a' prendre sur une case qui en contient *)

(* plusieurs choix : 

    1/ on a deja des paquetages, et il faut voir si on peut completer et comment?
       quitte a' en vider un peu 

    2/ on est vide 
       et on essaie de remplir de maniere optimale
       du genre on choisit le package de meilleur rapport poids/distance
       et on recupere ceux qui sont proches de celui-ci : 
          distance a' partir de ce point

*)

open State;;
open Distance;;  

type   pkg_infos = 
  {id : PackageSet.elt ; dest: State.dest; dist : int;  value : float};;

let gen_eval f d x = 
  let {x = px; y = py} as dest = Hashtbl.find  package_dests x in 
  let dt = get_distance d px py in 
  let v = f dest.weight dt in 
  {id = x; dest = dest; dist = dt; value = v};;

let zero = 0.00000;;

 
let check_zero f = (if abs_float(f) < zero then zero else f);;

let eval_neighbors = gen_eval (fun w d -> (float w) /. (4.0 *. (check_zero(float d))));; 

let count_cap l c  = 
  let rec count_rec l c r = match l with 
  [] -> r
  | t::q -> if t.dest.weight < c then (count_rec q (c-t.dest.weight) (t::r))
    else if t.dest.weight = c then t::r else (count_rec q c r)
  in 
  count_rec l c [];;
             
let neighbors eval_pkgl curr_cap = match eval_pkgl with 
  [] -> []
  | t::q ->
   let {x = x; y = y} = t.dest in
   let d = distance [(x, y)] in 
   let l1 =  List.map (fun x -> eval_neighbors d x.id) q in     
   let l2 =  List.filter (fun x ->  x.dest.weight < curr_cap) l1 in 
   t::(List.rev (count_cap l2 (curr_cap- t.dest.weight)));;
 


let eval_select  = gen_eval (fun w d -> (float w) /. (float d));;
 
let select local_point local_pkgl own_pkgl  own_cap   = 

  let d = distance [local_point] in 
  let eval_pkgl =  List.map (eval_select d)  (own_pkgl@local_pkgl)    in 
   let l1 =    List.filter (fun x ->  x.dest.weight <= own_cap)  eval_pkgl in 
   let l2 =    List.sort (fun x y -> if x.value <  y.value then 1 else if x = y then 0  else -1) l1 in 
    neighbors l2 own_cap ;;


  
