
(**
 $Id: select3.ml,v 1.3 2002/09/02 18:12:45 eleph Exp $
**)


(* choix d'un ensemble de package a' prendre sur une case qui en contient *)

(* plusieurs choix : 

    1/ on veut distribuer les packages que l'on a deja en esperant n'etre pas loin 
       a' la fin de la distribution d'un endroit ou' il y a des packages

       - soit il y a rien : on continue la distribution 
       - soit on tombe sur une case ou' il y a des packages : 

          a - on doit decider si on en prend, et si cela implique d'en laisser tomber 
              sachant que pour le calcul si on en depose on les depose tous 

              dans le cas ou'  on a rien par defaut on prend (quitte a' perdre un mouvement plus tard) 

       


*)

open State;;
open Distance;;  

type   pkg_infos = 
  {id : PackageSet.elt ; dest : State.dest ; dist : int;  value : float};;

let acces_dest pi = pi.dest;; 
let acces_dest_x pi = (acces_dest pi).x;;
let acces_dest_y pi = (acces_dest pi).y;;
let acces_dest_weight pi = (acces_dest pi).weight;;
let acces_dest_dt pi =  pi.dist;;

let epsilon = 0.001;;

let gen_eval f d x = 
  let { x = px ; y = py} as dest   = Hashtbl.find  package_dests x in 
  let dt = get_distance d px py in 
  let v =   f dest.weight dt in 
    {id = x; dest = dest ; dist = dt; value = v};;




let filter_select d max_money cap p = 
   let {x = px; y = py; weight = w} = p.dest in 
     (    ((get_distance d px py) <= max_money)  
       && (w <= cap))
;;


let eval_select  = gen_eval (fun w d -> if (abs_float (float d)) < epsilon then (float w) /. 0.5 else (float w) /. (float d));;

let same_position d1 d2 = 
   (d1.dest.x == d2.dest.x ) && (d1.dest.y == d2.dest.y);;

let split_hd h t = 
  let rec split_aux l l1 l2 = match l with 
    | [] -> l1,l2
    | nh::nt -> if same_position h nh then split_aux nt (nh::l1) l2 else split_aux nt l1 (nh::l2) 
  in split_aux t [] [];;

let rec regroupe l = match l with 
  |  [] -> [] 
  | h::t -> let hl,tl = split_hd h t   in (h::hl) :: (regroupe tl);;

let prepare l = 
  List.map (fun x -> let l_id = List.map (fun {id=i} -> i) x in 
            let pkg_info = (List.hd x)  in
	    let pos = acces_dest_x pkg_info, acces_dest_y pkg_info in  
	    l_id,pos) l
;;



let   select local_point local_money local_pkgl own_pkgl  own_cap   = 
  let all_pkgl = own_pkgl@local_pkgl in 
  let d = distance (local_point :: (List.map (fun z ->  z.dest.x,z.dest.y) own_pkgl)) in 

  let f_local_pkgl = List.filter (filter_select d local_money own_cap) local_pkgl in 

  let eval_pkgl =  List.map (eval_select d)  (List.map (fun x -> x.id) f_local_pkgl)    in 
  let l2 =    List.sort (fun x y -> if x.value -. y.value < 0.0 then -1 else 1) eval_pkgl  in 
    match l2 with 
      [] -> own_pkgl 
    | h::q -> let x = acces_dest_x h and y = acces_dest_y h in 
      let new_d = Distance.update_distance  d [(x,y)] in 
      let  l =  List.map (fun ip -> let x = acces_dest_x ip and  y = acces_dest_y ip in
		  let dt = Distance.get_distance new_d x y in  ip,dt) q  in 
      let ls = List.sort (fun x y -> (snd x) - (snd y) ) l in 
      let r = ref (own_cap  - (acces_dest_weight h)) in 
      let lid = 
        List.fold_left (fun res ip  -> let cap = acces_dest_weight ip in 
		                        if cap <= !r then ( r := !r - cap; ip :: res) else res) [h] q in
	

(*      let lpos = List.map (fun ip -> acces_dest_x ip, acces_dest_y ip) lid in *)
      let new_pkgl = lid@own_pkgl in new_pkgl ;;


  
		    

(*
select local_point (local_money)  q  (h::own_pkgl) (own_cap - h.dest.weight)
;;

*)

let main_select own_pkgl  = 
  let pos = State.my_location() in 
  let l_pkgl =  Hashtbl.find packages_at_location pos in 
  let fl_pkgl = List.map (fun (x,y) -> y ) !l_pkgl in 
  let l_dest = List.map (fun x -> x,(Hashtbl.find State.package_dests x)) fl_pkgl  in
  let local_pkgl  = List.map ( fun (x,d)  -> ( {id = x; dest = d;  dist = 0;  value =0.0}) ) l_dest  in
 

  let r = List.fold_left (fun x y -> x -y) !State.max_weight (List.map (fun x -> x.dest.weight) own_pkgl) in (* 30 in *)
  let new_pkgl = select pos !State.money local_pkgl own_pkgl r  in 
     let l1 = prepare(regroupe new_pkgl) in l1, [];;

(*
let init_eval l = st
  let d = distance [(State.my_location())] in 
  List.map 
    ( fun id -> 
        let dt = Hashtbl.find State.package_dests id in 
        let ds = get_distance d dt.x dt.y in 
        let v = eval_select dt.weight ds in  
                {id = id ; dest = dt; distance = ds; value = v }) l;;

*)




