
(* Number of moves *)
let moves = ref 0

(* Board description *)
let w = ref 0
let h = ref 0
let board = ref ""
let homes = ref ([] : (int * int) list)
let get_contents x y = !board.[x + y * !w]
let set_contents x y c = !board.[x + y * !w] <- c
let allowed_location x y = get_contents x y = '.'
let score = ref 0

(* Packages *)
type dest = {x : int; y : int; weight : int}
module PackageSet = Set.Make (struct type t = string let compare = compare end)
let package_list = ref PackageSet.empty
let package_locations : (string, [`Maybe|`Yes] * int * int) Hashtbl.t =
  Hashtbl.create 7
let package_dests : (string, dest) Hashtbl.t = Hashtbl.create 7
let package_carried : (string, string) Hashtbl.t = Hashtbl.create 7

let packages_at_location :
  (int * int, ([`Maybe|`Yes] * string) list ref) Hashtbl.t =
  Hashtbl.create 7

(* Robots *)
let robot_locations = Hashtbl.create 7
let robot_packages : (string, string list ref) Hashtbl.t = Hashtbl.create 7

(* Me *)
let my_id = ref ""
let max_weight = ref 0
let money = ref 0

let my_location () = Hashtbl.find robot_locations !my_id

(****)

let get_packages_at_location x y =
  try
    Hashtbl.find packages_at_location (x, y)
  with Not_found ->
    let l = ref [] in
    Hashtbl.replace packages_at_location (x, y) l;
    l

let remove_package id =
  begin try
    let (_, x, y) = Hashtbl.find package_locations id in
    let l = get_packages_at_location x y in
    l := List.filter (fun (_, id') -> id <> id') !l;
    if !l = [] then Hashtbl.remove packages_at_location (x, y)
  with Not_found -> () end;
  Hashtbl.remove package_locations id

let put_package id st x y =
  Hashtbl.replace package_locations id (st, x, y);
  let l = get_packages_at_location x y in
  l := (st, id) :: List.filter (fun (_, id') -> id <> id') !l

let kill_package id =
  remove_package id;
  Hashtbl.remove package_dests id;
  Hashtbl.remove package_carried id;
  package_list := PackageSet.remove id !package_list

let set_package id x y x' y' w =
  put_package id `Yes x' y';
  Hashtbl.replace package_dests id {x = x; y = y; weight = w};
  package_list := PackageSet.add id !package_list

let kill_unknown_packages x y =
  let l = !(get_packages_at_location x y) in
  List.iter
    (fun (st, id) ->
       match st with
         `Maybe -> kill_package id
       | _      -> ()) l

(****)

let get_robot_packages id =
  try
    Hashtbl.find robot_packages id
  with Not_found ->
    let l = ref [] in
    Hashtbl.replace robot_packages id l;
    l

let kill_robot id =
  Hashtbl.remove robot_locations id;
  let l = !(get_robot_packages id) in
  List.iter kill_package l;
  Hashtbl.remove robot_packages id

let set_robot id x y =
  if not (allowed_location x y) then
    kill_robot id
  else
    Hashtbl.replace robot_locations id (x, y)

let move_robot id dx dy =
  try
    let (x, y) = Hashtbl.find robot_locations id in
    set_robot id (x + dx) (y + dy)
  with Not_found ->
    Format.eprintf "Cannot move unknown robot %s@." id

let get_robot id =
  try
    Hashtbl.find robot_locations id
  with Not_found ->
    Format.eprintf "Unknown robot %s@." id;
    (0, 0)

let robot_pick_package id id' =
  let l = get_robot_packages id in
  l := id' :: !l;
  remove_package id';
  Hashtbl.replace package_carried id' id;
  package_list := PackageSet.add id' !package_list

let robot_drop_package id id' =
  (* We will not know all carried packages if we arrive when the game
     has already started *)
  package_list := PackageSet.add id' !package_list;
  let l = get_robot_packages id in
  l := List.filter (fun id'' -> id' <> id'') !l;
  if !l = [] then Hashtbl.remove robot_packages id;
  Hashtbl.remove package_carried id';
  let (x, y) = get_robot id in
  begin try
    let {x = x'; y = y'; weight = w} = Hashtbl.find package_dests id' in
    if x = x' && y = y' then begin
      if id = !my_id then score := !score + w;
      kill_package id'
    end else
      put_package id' `Yes x y
  with Not_found ->
    put_package id' `Maybe x y
  end
